#!/bin/env python
# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is mozilla.org code.
#
# The Initial Developer of the Original Code is
# The Mozilla Foundation
# Portions created by the Initial Developer are Copyright (C) 2007
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
# Ted Mielczarek <ted.mielczarek@gmail.com>
# Ben Turner <mozilla@songbirdnest.com>
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
#
# This version was forked from the mozilla-central version in 
# /toolkit/crashreporter/tools/ and changed to meet Camino's needs:
# - Fixes (hacks around) issues with file names/paths containing spaces.
# - Allows the contents of a dSYM bundle to be used directly if passed in
#   as one of the command-line arguments. (Note that it must be the symbol
#   file itself, not the containing .dSYM folder, or it will be skipped.)
# - Adds fall-back to non-DWARF symbols (see bug 493392).
# - When no actual symbols are generated for a binary, the useless file is
#   removed and its path is not echoed to stdout.
# - Commented out the 'subprocess' import, since that module isn't available
#   on Tiger. This will make --copy_debug fatal, but we don't currently use it.
# - Added a --no_dsym option that causes all symbols to be generated directly
#   from the binary (intended for use in generating OS symbols, where
#   trying to make a dSYM will fail due to permissions).
#
# Usage: symbolstore.py <params> <dump_syms path> <symbol store path>
#                                <debug info files or dirs>
#   Runs dump_syms on each debug info file specified on the command line,
#   then places the resulting symbol file in the proper directory
#   structure in the symbol store path.  Accepts multiple files
#   on the command line, so can be called as part of a pipe using
#   find <dir> | xargs symbolstore.pl <dump_syms> <storepath>
#   But really, you might just want to pass it <dir>.
#
#   Parameters accepted:
#     -c           : Copy debug info files to the same directory structure
#                    as sym files
#     -a "<archs>" : Run dump_syms -a <arch> for each space separated
#                    cpu architecture in <archs> (only on OS X)
#     -s <srcdir>  : Use <srcdir> as the top source directory to
#                    generate relative filenames.

import sys
import os
import re
import shutil
#from subprocess import call, STDOUT
from optparse import OptionParser

# Utility classes

class VCSFileInfo:
    """ A base class for version-controlled file information. Ensures that the
        following attributes are generated only once (successfully):

            self.root
            self.clean_root
            self.revision
            self.filename

        The attributes are generated by a single call to the GetRoot,
        GetRevision, and GetFilename methods. Those methods are explicitly not
        implemented here and must be implemented in derived classes. """

    def __init__(self, file):
        if not file:
            raise ValueError
        self.file = file

    def __getattr__(self, name):
        """ __getattr__ is only called for attributes that are not set on self,
            so setting self.[attr] will prevent future calls to the GetRoot,
            GetRevision, and GetFilename methods. We don't set the values on
            failure on the off chance that a future call might succeed. """

        if name == "root":
            root = self.GetRoot()
            if root:
                self.root = root
            return root

        elif name == "clean_root":
            clean_root = self.GetCleanRoot()
            if clean_root:
                self.clean_root = clean_root
            return clean_root

        elif name == "revision":
            revision = self.GetRevision()
            if revision:
                self.revision = revision
            return revision

        elif name == "filename":
            filename = self.GetFilename()
            if filename:
                self.filename = filename
            return filename

        raise AttributeError

    def GetRoot(self):
        """ This method should return the unmodified root for the file or 'None'
            on failure. """
        raise NotImplementedError

    def GetCleanRoot(self):
        """ This method should return the repository root for the file or 'None'
            on failure. """
        raise NotImplementedErrors

    def GetRevision(self):
        """ This method should return the revision number for the file or 'None'
            on failure. """
        raise NotImplementedError

    def GetFilename(self):
        """ This method should return the repository-specific filename for the
            file or 'None' on failure. """
        raise NotImplementedError

class CVSFileInfo(VCSFileInfo):
    """ A class to maintiain version information for files in a CVS repository.
        Derived from VCSFileInfo. """

    def __init__(self, file, srcdir):
        VCSFileInfo.__init__(self, file)
        self.srcdir = srcdir

    def GetRoot(self):
        (path, filename) = os.path.split(self.file)
        root = os.path.join(path, "CVS", "Root")
        if not os.path.isfile(root):
            return None
        f = open(root, "r")
        root_name = f.readline().strip()
        f.close()
        if root_name:
            return root_name
        print >> sys.stderr, "Failed to get CVS Root for %s" % filename
        return None

    def GetCleanRoot(self):
        parts = self.root.split('@')
        if len(parts) > 1:
            # we don't want the extra colon
            return parts[1].replace(":","")
        return self.root.replace(":","")

    def GetRevision(self):
        (path, filename) = os.path.split(self.file)
        entries = os.path.join(path, "CVS", "Entries")
        if not os.path.isfile(entries):
            return None
        f = open(entries, "r")
        for line in f:
            parts = line.split("/")
            if len(parts) > 1 and parts[1] == filename:
                return parts[2]
        print >> sys.stderr, "Failed to get CVS Revision for %s" % filename
        return None

    def GetFilename(self):
        file = self.file
        if self.revision and self.clean_root:
            if self.srcdir:
                # strip the base path off
                # but we actually want the last dir in srcdir
                file = os.path.normpath(file)
                # the lower() is to handle win32+vc8, where
                # the source filenames come out all lowercase,
                # but the srcdir can be mixed case
                if file.lower().startswith(self.srcdir.lower()):
                    file = file[len(self.srcdir):]
                (head, tail) = os.path.split(self.srcdir)
                if tail == "":
                    tail = os.path.basename(head)
                file = tail + file
            return "cvs:%s:%s:%s" % (self.clean_root, file, self.revision)
        return file

# This regex separates protocol and optional username/password from a url.
# For instance, all the following urls will be transformed into
# 'foo.com/bar':
#
#   http://foo.com/bar
#   svn+ssh://user@foo.com/bar
#   svn+ssh://user:pass@foo.com/bar
#
# This is used by both SVN and HG
rootRegex = re.compile(r'^\S+?:/+(?:[^\s/]*@)?(\S+)$')

class SVNFileInfo(VCSFileInfo):
    url = None
    repo = None
    svndata = {}

    def __init__(self, file):
        """ We only want to run subversion's info tool once so pull all the data
            here. """

        VCSFileInfo.__init__(self, file)

        if os.path.isfile(file):
            command = os.popen("svn info %s" % file, "r")
            for line in command:
                # The last line of the output is usually '\n'
                if line.strip() == '':
                    continue
                # Split into a key/value pair on the first colon
                key, value = line.split(':', 1)
                if key in ["Repository Root", "Revision", "URL"]:
                    self.svndata[key] = value.strip()

            exitStatus = command.close()
            if exitStatus:
              print >> sys.stderr, "Failed to get SVN info for %s" % file

    def GetRoot(self):
        key = "Repository Root"
        if key in self.svndata:
            match = rootRegex.match(self.svndata[key])
            if match:
                return match.group(1)
        print >> sys.stderr, "Failed to get SVN Root for %s" % self.file
        return None

    # File bug to get this teased out from the current GetRoot, this is temporary
    def GetCleanRoot(self):
        return self.root

    def GetRevision(self):
        key = "Revision"
        if key in self.svndata:
            return self.svndata[key]
        print >> sys.stderr, "Failed to get SVN Revision for %s" % self.file
        return None

    def GetFilename(self):
        if self.root and self.revision:
            if "URL" in self.svndata and "Repository Root" in self.svndata:
                url, repo = self.svndata["URL"], self.svndata["Repository Root"]
                file = url[len(repo) + 1:]
            return "svn:%s:%s:%s" % (self.root, file, self.revision)
        print >> sys.stderr, "Failed to get SVN Filename for %s" % self.file
        return self.file

class HGRepoInfo:
    # HG info is per-repo, so cache it in a static
    # member var
    repos = {}
    def __init__(self, path, rev, cleanroot):
        self.path = path
        self.rev = rev
        self.cleanroot = cleanroot

class HGFileInfo(VCSFileInfo):
    def __init__(self, file, srcdir):
        VCSFileInfo.__init__(self, file)
        # we should only have to collect this info once per-repo
        if not srcdir in HGRepoInfo.repos:
            rev = os.popen('hg identify -i "%s"' % srcdir, "r").readlines()[0].rstrip()
            # could have a + if there are uncommitted local changes
            if rev.endswith('+'):
                rev = rev[:-1]

            path = os.popen('hg -R "%s" showconfig paths.default' % srcdir, "r").readlines()[0].rstrip()
            if path == '':
                hg_root = os.environ.get("SRCSRV_ROOT")
                if hg_root:
                    path = hg_root
                else:
                    print >> sys.stderr, "Failed to get HG Repo for %s" % srcdir
            if path != '': # not there?
                match = rootRegex.match(path)
                if match:
                    cleanroot = match.group(1)
                    if cleanroot.endswith('/'):
                        cleanroot = cleanroot[:-1]
            HGRepoInfo.repos[srcdir] = HGRepoInfo(path, rev, cleanroot)
        self.repo = HGRepoInfo.repos[srcdir]
        self.file = file
        self.srcdir = srcdir

    def GetRoot(self):
        return self.repo.path

    def GetCleanRoot(self):
        return self.repo.cleanroot

    def GetRevision(self):
        return self.repo.rev

    def GetFilename(self):
        file = self.file
        if self.revision and self.clean_root:
            if self.srcdir:
                # strip the base path off
                file = os.path.normpath(file)
                if IsInDir(file, self.srcdir):
                    file = file[len(self.srcdir):]
                if file.startswith('/') or file.startswith('\\'):
                    file = file[1:]
            return "hg:%s:%s:%s" % (self.clean_root, file, self.revision)
        return file

# Utility functions

# A cache of files for which VCS info has already been determined. Used to
# prevent extra filesystem activity or process launching.
vcsFileInfoCache = {}

def IsInDir(file, dir):
    # the lower() is to handle win32+vc8, where
    # the source filenames come out all lowercase,
    # but the srcdir can be mixed case
    return os.path.abspath(file).lower().startswith(os.path.abspath(dir).lower())

def GetVCSFilename(file, srcdirs):
    """Given a full path to a file, and the top source directory,
    look for version control information about this file, and return
    a tuple containing
    1) a specially formatted filename that contains the VCS type,
    VCS location, relative filename, and revision number, formatted like:
    vcs:vcs location:filename:revision
    For example:
    cvs:cvs.mozilla.org/cvsroot:mozilla/browser/app/nsBrowserApp.cpp:1.36
    2) the unmodified root information if it exists"""
    (path, filename) = os.path.split(file)
    if path == '' or filename == '':
        return (file, None)

    fileInfo = None
    root = ''
    if file in vcsFileInfoCache:
        # Already cached this info, use it.
        fileInfo = vcsFileInfoCache[file]
    else:
        for srcdir in srcdirs:
            if os.path.isdir(os.path.join(path, "CVS")):
                fileInfo = CVSFileInfo(file, srcdir)
                if fileInfo:
                    root = fileInfo.root
            elif os.path.isdir(os.path.join(path, ".svn")) or \
                 os.path.isdir(os.path.join(path, "_svn")):
                 fileInfo = SVNFileInfo(file);
            elif os.path.isdir(os.path.join(srcdir, '.hg')) and \
                 IsInDir(file, srcdir):
                 fileInfo = HGFileInfo(file, srcdir)

            if fileInfo: 
                vcsFileInfoCache[file] = fileInfo
                break

    if fileInfo:
        file = fileInfo.filename
        root = fileInfo.root

    # we want forward slashes on win32 paths
    return (file.replace("\\", "/"), root)

def GetPlatformSpecificDumper(**kwargs):
    """This function simply returns a instance of a subclass of Dumper
    that is appropriate for the current platform."""
    return {'win32': Dumper_Win32,
            'cygwin': Dumper_Win32,
            'linux2': Dumper_Linux,
            'sunos5': Dumper_Solaris,
            'darwin': Dumper_Mac}[sys.platform](**kwargs)

def SourceIndex(fileStream, outputPath, vcs_root):
    """Takes a list of files, writes info to a data block in a .stream file"""
    # Creates a .pdb.stream file in the mozilla\objdir to be used for source indexing
    # Create the srcsrv data block that indexes the pdb file
    result = True
    pdbStreamFile = open(outputPath, "w")
    pdbStreamFile.write('''SRCSRV: ini ------------------------------------------------\r\nVERSION=2\r\nINDEXVERSION=2\r\nVERCTRL=http\r\nSRCSRV: variables ------------------------------------------\r\nHGSERVER=''')
    pdbStreamFile.write(vcs_root)
    pdbStreamFile.write('''\r\nSRCSRVVERCTRL=http\r\nHTTP_EXTRACT_TARGET=%hgserver%/raw-file/%var3%/%var2%\r\nSRCSRVTRG=%http_extract_target%\r\nSRCSRV: source files ---------------------------------------\r\n''')
    pdbStreamFile.write(fileStream) # can't do string interpolation because the source server also uses this and so there are % in the above
    pdbStreamFile.write("SRCSRV: end ------------------------------------------------\r\n\n")
    pdbStreamFile.close()
    return result

class Dumper:
    """This class can dump symbols from a file with debug info, and
    store the output in a directory structure that is valid for use as
    a Breakpad symbol server.  Requires a path to a dump_syms binary--
    |dump_syms| and a directory to store symbols in--|symbol_path|.
    Optionally takes a list of processor architectures to process from
    each debug file--|archs|, the full path to the top source
    directory--|srcdir|, for generating relative source file names,
    and an option to copy debug info files alongside the dumped
    symbol files--|copy_debug|, mostly useful for creating a
    Microsoft Symbol Server from the resulting output.

    You don't want to use this directly if you intend to call
    ProcessDir.  Instead, call GetPlatformSpecificDumper to
    get an instance of a subclass."""
    def __init__(self, dump_syms, symbol_path,
                 archs=None, srcdirs=None, copy_debug=False, vcsinfo=False, srcsrv=False, no_dsym=False):
        # popen likes absolute paths, at least on windows
        self.dump_syms = os.path.abspath(dump_syms)
        self.symbol_path = symbol_path
        if archs is None:
            # makes the loop logic simpler
            self.archs = ['']
        else:
            self.archs = ['-a %s' % a for a in archs.split()]
        if srcdirs is not None:
            self.srcdirs = [os.path.normpath(a) for a in srcdirs]
        else:
            self.srcdirs = None
        self.copy_debug = copy_debug
        self.vcsinfo = vcsinfo
        self.srcsrv = srcsrv
        self.no_dsym = no_dsym

    # subclasses override this
    def ShouldProcess(self, file):
        return False

    # and can override this
    def ShouldSkipDir(self, dir):
        return False

    def RunFileCommand(self, file):
        """Utility function, returns the output of file(1)"""
        try:
            # we use -L to read the targets of symlinks,
            # and -b to print just the content, not the filename
            return os.popen("file -Lb \"" + file + "\"").read()
        except:
            return ""

    # This is a no-op except on Win32
    def FixFilenameCase(self, file):
        return file

    # This is a no-op except on Win32
    def SourceServerIndexing(self, debug_file, guid, sourceFileStream, vcs_root):
        return ""

    # subclasses override this if they want to support this
    def CopyDebug(self, file, debug_file, guid):
        pass

    def Process(self, file_or_dir):
        "Process a file or all the (valid) files in a directory."
        if os.path.isdir(file_or_dir) and not self.ShouldSkipDir(file_or_dir):
            return self.ProcessDir(file_or_dir)
        elif os.path.isfile(file_or_dir):
            return self.ProcessFile(file_or_dir)
        # maybe it doesn't exist?
        return False

    def ProcessDir(self, dir):
        """Process all the valid files in this directory.  Valid files
        are determined by calling ShouldProcess."""
        result = True
        for root, dirs, files in os.walk(dir):
            for d in dirs[:]:
                if self.ShouldSkipDir(d):
                    dirs.remove(d)
            for f in files:
                fullpath = os.path.join(root, f)
                if self.ShouldProcess(fullpath):
                    if not self.ProcessFile(fullpath):
                        result = False
        return result

    def ProcessFile(self, file):
        """Dump symbols from this file into a symbol file, stored
        in the proper directory structure in  |symbol_path|."""
        processedLines = 0
        sourceFileStream = ''
        # tries to get the vcs root from the .mozconfig first - if it's not set
        # the tinderbox vcs path will be assigned further down
        vcs_root = os.environ.get("SRCSRV_ROOT")
        for arch in self.archs:
            try:
                cmd = os.popen("%s %s \"%s\"" % (self.dump_syms, arch, file), "r")
                module_line = cmd.next()
                if module_line.startswith("MODULE"):
                    # MODULE os cpu guid debug_file
                    (guid, debug_file) = (module_line.split(None, 4))[3:5]
                    if guid == "000000000000000000000000000000000":
                        continue
                    debug_file = debug_file.rstrip()
                    # strip off .pdb extensions, and append .sym
                    sym_file = re.sub("\.pdb$", "", debug_file) + ".sym"
                    # we do want forward slashes here
                    rel_path = os.path.join(debug_file,
                                            guid,
                                            sym_file).replace("\\", "/")
                    full_path = os.path.normpath(os.path.join(self.symbol_path,
                                                              rel_path))
                    try:
                        os.makedirs(os.path.dirname(full_path))
                    except OSError: # already exists
                        pass
                    f = open(full_path, "w")
                    f.write(module_line)
                    # now process the rest of the output
                    for line in cmd:
                        if not line.startswith("MODULE"):
                            processedLines += 1
                        if line.startswith("FILE"):
                            # FILE index filename
                            (x, index, filename) = line.split(None, 2)
                            if sys.platform == "sunos5":
                                for srcdir in self.srcdirs:
                                    start = filename.find(self.srcdir)
                                    if start != -1:
                                        filename = filename[start:]
                                        break
                            filename = self.FixFilenameCase(filename.rstrip())
                            sourcepath = filename
                            if self.vcsinfo:
                                (filename, rootname) = GetVCSFilename(filename, self.srcdirs)
                                # sets vcs_root in case the loop through files were to end on an empty rootname
                                if vcs_root is None:
                                  if rootname:
                                     vcs_root = rootname
                            # gather up files with hg for indexing   
                            if filename.startswith("hg"):
                                (ver, checkout, source_file, revision) = filename.split(":", 3)
                                sourceFileStream += sourcepath + "*" + source_file + '*' + revision + "\r\n"
                            f.write("FILE %s %s\n" % (index, filename))
                        else:
                            # pass through all other lines unchanged
                            f.write(line)
                    f.close()
                    cmd.close()
                    # If the symbol file was empty, throw it away and don't report it.
                    if processedLines == 0:
                        os.remove(full_path)
                        # If the two enclosing directories that we (potentially) made above
                        # are empty, remove them too. If they aren't empty, fail silently.
                        parent_dir = os.path.dirname(full_path)
                        try:
                            os.rmdir(parent_dir)
                            os.rmdir(os.path.dirname(parent_dir))
                        except:
                            pass
                        continue
                    # we output relative paths so callers can get a list of what
                    # was generated
                    print rel_path
                    if self.copy_debug:
                        self.CopyDebug(file, debug_file, guid)
                    if self.srcsrv and vcs_root:
                        # Call on SourceServerIndexing
                        result = self.SourceServerIndexing(debug_file, guid, sourceFileStream, vcs_root)
            except StopIteration:
                pass
            except:
                print >> sys.stderr, "Unexpected error: ", sys.exc_info()[0]
                raise
        return (processedLines > 0)

# Platform-specific subclasses.  For the most part, these just have
# logic to determine what files to extract symbols from.

class Dumper_Win32(Dumper):
    fixedFilenameCaseCache = {}

    def ShouldProcess(self, file):
        """This function will allow processing of pdb files that have dll
        or exe files with the same base name next to them."""
        if file.endswith(".pdb"):
            (path,ext) = os.path.splitext(file)
            if os.path.isfile(path + ".exe") or os.path.isfile(path + ".dll"):
                return True
        return False

    def FixFilenameCase(self, file):
        """Recent versions of Visual C++ put filenames into
        PDB files as all lowercase.  If the file exists
        on the local filesystem, fix it."""

        # Use a cached version if we have one.
        if file in self.fixedFilenameCaseCache:
            return self.fixedFilenameCaseCache[file]

        result = file

        (path, filename) = os.path.split(file)
        if os.path.isdir(path):
            lc_filename = filename.lower()
            for f in os.listdir(path):
                if f.lower() == lc_filename:
                    result = os.path.join(path, f)
                    break

        # Cache the corrected version to avoid future filesystem hits.
        self.fixedFilenameCaseCache[file] = result
        return result

    def CopyDebug(self, file, debug_file, guid):
        rel_path = os.path.join(debug_file,
                                guid,
                                debug_file).replace("\\", "/")
        full_path = os.path.normpath(os.path.join(self.symbol_path,
                                                  rel_path))
        shutil.copyfile(file, full_path)
        # try compressing it
        compressed_file = os.path.splitext(full_path)[0] + ".pd_"
        # ignore makecab's output
        success = call(["makecab.exe", full_path, compressed_file],
                       stdout=open("NUL:","w"), stderr=STDOUT)
        if success == 0 and os.path.exists(compressed_file):
            os.unlink(full_path)
            print os.path.splitext(rel_path)[0] + ".pd_"
        else:
            print rel_path
        
    def SourceServerIndexing(self, debug_file, guid, sourceFileStream, vcs_root):
        # Creates a .pdb.stream file in the mozilla\objdir to be used for source indexing
        cwd = os.getcwd()
        streamFilename = debug_file + ".stream"
        stream_output_path = os.path.join(cwd, streamFilename)
        # Call SourceIndex to create the .stream file
        result = SourceIndex(sourceFileStream, stream_output_path, vcs_root)
        
        if self.copy_debug:
            pdbstr_path = os.environ.get("PDBSTR_PATH")
            pdbstr = os.path.normpath(pdbstr_path)
            pdb_rel_path = os.path.join(debug_file, guid, debug_file)
            pdb_filename = os.path.normpath(os.path.join(self.symbol_path, pdb_rel_path))
            # move to the dir with the stream files to call pdbstr
            os.chdir(os.path.dirname(stream_output_path))
            os.spawnv(os.P_WAIT, pdbstr, [pdbstr, "-w", "-p:" + pdb_filename, "-i:" + streamFilename, "-s:srcsrv"])
            # clean up all the .stream files when done
            os.remove(stream_output_path)
        return result

class Dumper_Linux(Dumper):
    def ShouldProcess(self, file):
        """This function will allow processing of files that are
        executable, or end with the .so extension, and additionally
        file(1) reports as being ELF files.  It expects to find the file
        command in PATH."""
        if file.endswith(".so") or os.access(file, os.X_OK):
            return self.RunFileCommand(file).startswith("ELF")
        return False

    def CopyDebug(self, file, debug_file, guid):
        # We want to strip out the debug info, and add a
        # .gnu_debuglink section to the object, so the debugger can
        # actually load our debug info later.
        file_dbg = file + ".dbg"
        os.system("objcopy --only-keep-debug %s %s" % (file, file_dbg))
        os.system("objcopy --add-gnu-debuglink=%s %s" % (file_dbg, file))
        
        rel_path = os.path.join(debug_file,
                                guid,
                                debug_file + ".dbg")
        full_path = os.path.normpath(os.path.join(self.symbol_path,
                                                  rel_path))
        shutil.move(file_dbg, full_path)
        # gzip the shipped debug files
        os.system("gzip %s" % full_path)
        print rel_path + ".gz"

class Dumper_Solaris(Dumper):
    def RunFileCommand(self, file):
        """Utility function, returns the output of file(1)"""
        try:
            output = os.popen("file " + file).read()
            return output.split('\t')[1];
        except:
            return ""

    def ShouldProcess(self, file):
        """This function will allow processing of files that are
        executable, or end with the .so extension, and additionally
        file(1) reports as being ELF files.  It expects to find the file
        command in PATH."""
        if file.endswith(".so") or os.access(file, os.X_OK):
            return self.RunFileCommand(file).startswith("ELF")
        return False

class Dumper_Mac(Dumper):
    def ShouldProcess(self, file):
        """This function will allow processing of files that are
        executable, or end with the .dylib extension, and additionally
        file(1) reports as being Mach-O files.  It expects to find the file
        command in PATH."""
        if file.endswith(".dylib") or os.access(file, os.X_OK):
            return self.RunFileCommand(file).startswith("Mach-O")
        return False

    def ShouldSkipDir(self, dir):
        """We create .dSYM bundles on the fly, but if someone runs
        buildsymbols twice, we should skip any bundles we created
        previously, otherwise we'll recurse into them and try to 
        dump the inner bits again."""
        if dir.endswith(".dSYM"):
            return True
        return False

    def ProcessFile(self, file):
        """dump_syms on Mac needs to be run on a dSYM bundle produced
        by dsymutil(1), so run dsymutil here and pass the bundle name
        down to the superclass method instead."""
        res = False
        # If there's a dSYM in the path then we must have been given an
        # existing dSYM file as an argument (since ShouldSkipDir would have
        # skipped it), so we should use it directly.
        # If no_dsym is set, then everything bypasses the dSYM generation.
        if self.no_dsym or file.find('.dSYM/') != -1:
            res = Dumper.ProcessFile(self, file)
        else:
            dsymbundle = file + ".dSYM"
            if os.path.exists(dsymbundle):
                shutil.rmtree(dsymbundle)
            # dsymutil takes --arch=foo instead of -a foo like everything else
            os.system("dsymutil %s \"%s\" >/dev/null" % (' '.join([a.replace('-a ', '--arch=') for a in self.archs]),
                                      file))
            res = Dumper.ProcessFile(self, dsymbundle)
            # CopyDebug will already have been run from Dumper.ProcessFile
            shutil.rmtree(dsymbundle)
            # If there were no DWARF symbols, run dump_syms on the binary
            # itself so we can at least get minimal symbol information.
            if not res:
                print >> sys.stderr, "No DWARF symbols for %s; using fallback symbols" % file
                res = Dumper.ProcessFile(self, file)
        return res

    def CopyDebug(self, file, debug_file, guid):
        """ProcessFile has already produced a dSYM bundle, so we should just
        copy that to the destination directory. However, we'll package it
        into a .tar.bz2 because the debug symbols are pretty huge, and
        also because it's a bundle, so it's a directory. |file| here is the
        dSYM bundle, and |debug_file| is the original filename."""
        rel_path = os.path.join(debug_file,
                                guid,
                                os.path.basename(file) + ".tar.bz2")
        full_path = os.path.abspath(os.path.join(self.symbol_path,
                                                  rel_path))
        success = call(["tar", "cjf", full_path, os.path.basename(file)],
                       cwd=os.path.dirname(file),
                       stdout=open("/dev/null","w"), stderr=STDOUT)
        if success == 0 and os.path.exists(full_path):
            print rel_path

# Entry point if called as a standalone program
def main():
    parser = OptionParser(usage="usage: %prog [options] <dump_syms binary> <symbol store path> <debug info files>")
    parser.add_option("-c", "--copy",
                      action="store_true", dest="copy_debug", default=False,
                      help="Copy debug info files into the same directory structure as symbol files")
    parser.add_option("-a", "--archs",
                      action="store", dest="archs",
                      help="Run dump_syms -a <arch> for each space separated cpu architecture in ARCHS (only on OS X)")
    parser.add_option("-s", "--srcdir",
                      action="append", dest="srcdir", default=[],
                      help="Use SRCDIR to determine relative paths to source files")
    parser.add_option("-v", "--vcs-info",
                      action="store_true", dest="vcsinfo",
                      help="Try to retrieve VCS info for each FILE listed in the output")
    parser.add_option("-i", "--source-index",
                      action="store_true", dest="srcsrv", default=False,
                      help="Add source index information to debug files, making them suitable for use in a source server.")
    parser.add_option("--no-dsym",
                      action="store_true", dest="no_dsym", default=False,
                      help="Runs dumpsyms directly on the binary, rather than attempting to generate a dSYM bundle.")
    (options, args) = parser.parse_args()
    
    #check to see if the pdbstr.exe exists
    if options.srcsrv:
        pdbstr = os.environ.get("PDBSTR_PATH")
        if not os.path.exists(pdbstr):
            print >> sys.stderr, "Invalid path to pdbstr.exe - please set/check PDBSTR_PATH.\n"
            sys.exit(1)
            
    if len(args) < 3:
        parser.error("not enough arguments")
        exit(1)

    dumper = GetPlatformSpecificDumper(dump_syms=args[0],
                                       symbol_path=args[1],
                                       copy_debug=options.copy_debug,
                                       archs=options.archs,
                                       srcdirs=options.srcdir,
                                       vcsinfo=options.vcsinfo,
                                       srcsrv=options.srcsrv,
                                       no_dsym=options.no_dsym)
    for arg in args[2:]:
        dumper.Process(arg)

# run main if run directly
if __name__ == "__main__":
    main()



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is mozilla.org code
#
# The Initial Developer of the Original Code is mozilla.org.
# Portions created by the Initial Developer are Copyright (C) 2005
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Mark Hammond: initial author
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

# A utility for compiling Python code, using features not available via
# the builtin compile.
#
# (a) It is not possible to compile the body of a Python function, without the
# function declaration. ie, 'return None' will always give a syntax error when
# passed to compile.
# (b) It is very tricky to compile code with the line-number starting at
# anything other than zero.
#
# Both of these are solved by this module, which uses the 'compiler' module
# XXX - sad side-effect is that Unicode is not correctly supported - 
# PyCF_SOURCE_IS_UTF8 is not exposed via compiler (in 2.3 at least)
# On the upside here, all 'src' params are unicode today, so expansion here
# requires no interface changes.

from compiler import parse, syntax, compile
from compiler.pycodegen import ModuleCodeGenerator
import compiler.ast
import new

def _fix_src(src):
    # windows first - \r\n -> \n, then for mac, remaining \r -> \n
    # Trailing whitespace can cause problems - make sure a final '\n' exists.
    return src.replace("\r\n", "\n").replace("\r", "\n") + "\n"

# from compiler.misc.set_filename - but we also adjust lineno attributes.
def set_filename_and_offset(filename, offset, tree):
    """Set the filename attribute to filename on every node in tree"""
    worklist = [tree]
    while worklist:
        node = worklist.pop(0)
        node.filename = filename
        if node.lineno is not None:
            node.lineno += offset
        worklist.extend(node.getChildNodes())

def parse_function(src, func_name, arg_names, defaults=[]):
    tree = parse(src, "exec")
    defaults = [compiler.ast.Const(d) for d in defaults]
    # Insert a Stmt with function object.
    try:
        decs = compiler.ast.Decorators([])
    except AttributeError:
        # 2.3 has no such concept (and different args!)
        func = compiler.ast.Function(func_name, arg_names, defaults, 0, None,
                                     tree.node)
    else:
        # 2.4 and later
        func = compiler.ast.Function(decs, func_name, arg_names, defaults, 0, None,
                                     tree.node)
    stmt = compiler.ast.Stmt((func,))
    tree.node = stmt
    syntax.check(tree)
    return tree

def compile_function(src, filename, func_name, arg_names, defaults=[],
                     # more args to come...
                     lineno=0): 
    assert filename, "filename is required"
    try:
        tree = parse_function(_fix_src(src), func_name, arg_names, defaults)
    except SyntaxError, err:
        err.lineno += lineno
        err.filename = filename
        raise SyntaxError, err

    set_filename_and_offset(filename, lineno, tree)

    gen = ModuleCodeGenerator(tree)
    return gen.getCode()

# And a 'standard' compile, but with the filename offset feature.
def compile(src, filename, mode='exec', flags=None, dont_inherit=None, lineno=0):
    if flags is not None or dont_inherit is not None or mode != 'exec':
        raise RuntimeError, "not implemented yet"
    try:
        tree = parse(_fix_src(src), mode)
    except SyntaxError, err:
        err.lineno += lineno
        err.filename = filename
        raise SyntaxError, err

    set_filename_and_offset(filename, lineno, tree)

    gen = ModuleCodeGenerator(tree)
    return gen.getCode()



# This is vaguely related to an nsIScriptContext.  The pydom C code
# just calls back into this class - but that C code still retains some of the
# functionality.

import sys, types, new, logging

import domcompile

from xpcom.client import Component
from xpcom import components, primitives, COMException, nsError, _xpcom

import _nsdom

GlobalWindowIIDs = [_nsdom.IID_nsIScriptGlobalObject,
                    components.interfaces.nsIDOMWindow]

IID_nsIDOMXULElement = components.interfaces.nsIDOMXULElement

# Borrow the xpcom logger (but use a new sub-logger)
logger = logging.getLogger("xpcom.nsdom")
# Note that many calls to logger.debug are prefixed with if __debug__.
# This is to prevent *any* penalty in non-debug mode for logging in perf
# critical areas.

# Also note that assert statements have no cost in non-debug builds.

# XUL Elements don't expose the XBL implemented interfaces via classinfo.
# So we cheat :)
# Most support nsIAccessibleProvider - don't list that!
# Although this sucks for now, if you strike something that isn't listed,
# then simply QI the object you have for the interface you need to use. To
# avoid the explicit QI, you could also have your code do:
#    import nsdom.context
#    xei = nsdom.context.xul_element_interfaces
#    # flag for manual inspection should a future version include it
#    if 'somewidget' in xei: raise RuntimeError, "already here!?"
#    xei['somewidget'] = [...]

xul_element_interfaces = {
    # tagName:      [IID, ...]
    'textbox':      [components.interfaces.nsIDOMXULTextBoxElement],
    'button':       [components.interfaces.nsIDOMXULButtonElement],
    'checkbox':     [components.interfaces.nsIDOMXULCheckboxElement],
    'image':        [components.interfaces.nsIDOMXULImageElement],
    'tree':         [components.interfaces.nsIDOMXULTreeElement,
                     components.interfaces.nsIDOMXULMultiSelectControlElement],
    'listbox':      [components.interfaces.nsIDOMXULMultiSelectControlElement],
    'menu':         [components.interfaces.nsIDOMXULSelectControlItemElement],
    'popup':        [components.interfaces.nsIDOMXULPopupElement],
    'radiogroup':   [components.interfaces.nsIDOMXULSelectControlElement],
}

def dump(fmt, *args):
    """A global 'dump' function, available in global namespaces.  Similar to
    The JS version, but this supports % message formatting.

    Main advantage over 'print' is that it catches the possible IO error
    when there is no available console.
    """
    try:
        sys.stderr.write(fmt % args)
        sys.stderr.write("\n")
    except IOError:
        pass

# The event listener class we attach to an object for addEventListener
class EventListener:
    _com_interfaces_ = components.interfaces.nsIDOMEventListener
    def __init__(self, context, evt_name, handler, globs):
        # 'globs' are the globals the event handler will use, or None
        # if the handler should use the globals it was compiled with.
        self.context = context
        if callable(handler):
            self.func = handler
        else:
            # event name for addEventListener has no 'on' prefix
            func_name = "on" + evt_name
            # What to do about arg names?  It looks like onerror
            # still only gets one arg here anyway - handleEvent only takes
            # one!
            arg_names = ('event',)
            co = domcompile.compile_function(handler,
                                             "inline event '%s'" % evt_name,
                                             func_name,
                                             arg_names)
            # Could use globs here, but should not be necessary - all we
            # are doing is getting the function object from it.
            g = {}
            exec co in g
            self.func = g[func_name]
        self.globals = globs

    def handleEvent(self, event):
        # Although handler is already a function object, we must re-bind to
        # new globals if necessary.
        if self.globals is not None:
            f = new.function(self.func.func_code, self.globals, 
                            self.func.func_name, self.func.func_closure)
        else:
            f = self.func
        # Convert the raw pyxpcom object to a magic _nsdom one, that
        # knows how to remember expandos etc based on context (ie, winds
        # up calling back into MakeInterfaceResult().
        event = _nsdom.MakeDOMObject(self.context, event)
        
        args = (event,)
        # We support having less args declared than supplied, a-la JS.
        # (This can only happen when passed a function object - we always
        # compile a string handler into a function with 1 arg)
        args = args[:f.func_code.co_argcount]
        return f(*args)

class WrappedNative(Component):
    """Implements the xpconnect concept of 'wrapped natives' and 'expandos'.

    DOM objects can have arbitrary values set on them.  Once this is done for
    the first time, it gets stored in a map in the context.  This leads to
    cycles, which must be cleaned up when the context is closed.
    """
    def __init__(self, context, obj, iid):
        # Store our context - but our context doesn't keep a reference
        # to us until someone calls self._remember_object() on the context -
        # which sets up all kinds of cycles!
        self.__dict__['_context_'] = context
        # We store expandos in a separate dict rather than directly in our
        # __dict__.  No real need for this other than to prevent these
        # attributes clobbering ones we need to work!
        self.__dict__['_expandos_'] = {}
        Component.__init__(self, obj, iid)

    def __repr__(self):
        iface_desc = self._get_classinfo_repr_()
        return "<DOM object '%s' (%s)>" % (self._object_name_,iface_desc)
        
    def __getattr__(self, attr):
        # If it exists in expandos, always return it.
        if attr.startswith("__"):
            raise AttributeError, attr
        # expandos come before interface attributes (which may be wrong.  If
        # we do this, why not just store it in __dict__?)
        expandos = self._expandos_
        if expandos.has_key(attr):
            return expandos[attr]
        return Component.__getattr__(self, attr)

    def __setattr__(self, attr, value):
        try:
            Component.__setattr__(self, attr, value)
        except AttributeError:
            # Set it as an 'expando'.  It looks like we should delegate *all*
            # to the outside object.
            logger.debug("setting expando %s.%r=%r for object %r",
                         self, attr, value, self._comobj_)
            # and register if an event.
            if attr.startswith("on"):
                # I'm quite confused by this :(
                target = self._comobj_
                if _nsdom.IsOuterWindow(target):
                    target = _nsdom.GetCurrentInnerWindow(target)
                go = self._context_.globalObject
                scope = self._context_.GetNativeGlobal()
                if callable(value):
                    # no idea if this is right - set the compiled object ourselves.
                    self._expandos_[attr] = value
                    _nsdom.RegisterScriptEventListener(go, scope, target, attr)
                else:
                    _nsdom.AddScriptEventListener(target, attr, value, False, False)
                    _nsdom.CompileScriptEventListener(go, scope, target, attr)
            else:
                self._expandos_[attr] = value
            self._context_._remember_object(self)

    def _build_all_supported_interfaces_(self):
        # Generally called as pyxpcom is finding an attribute, overridden to
        # work around lack of class info for xbl bindings.
        Component._build_all_supported_interfaces_(self)
        # Now hard-code certain element names we know about, as the XBL
        # implemented interfaces are not exposed by this.
        ii = self.__dict__['_interface_infos_']
        if ii.has_key(IID_nsIDOMXULElement):
            # Is a DOM element - see if in our map.
            tagName = self.tagName
            interfaces = xul_element_interfaces.get(tagName, [])
            for interface in interfaces:
                if not ii.has_key(interface):
                        self._remember_interface_info(interface)
        else:
            logger.info("Not a DOM element - not hooking extra interfaces")

    def addEventListener(self, event, handler, useCapture=False, globs=None):
        # We need to transform string or function objects into
        # nsIDOMEventListener interfaces.
        logger.debug("addEventListener for %r, event=%r, handler=%r, cap=%s",
                     self, event, handler, useCapture)

        if not hasattr(handler, "handleEvent"): # may already be a handler instance.
            # Wrap it in our instance, which knows how to convert strings and
            # function objects.

            # Handle semantics for event handler globals:
            # * If a function, then function globals are used, as per normal.
            # * If a string object and no explicit globals param, use the
            #   caller's globals
            if globs is None and not callable(handler):
                globs = sys._getframe().f_back.f_globals
            handler = EventListener(self._context_, event, handler, globs)
        else:
            assert not globs, "can't specify a handler instance and globals"

        base = self.__getattr__('addEventListener')
        base(event, handler, useCapture)

class WrappedNativeGlobal(WrappedNative):
    # Special support for our global object.  Certain methods exposed by
    # IDL are JS specific - generally ones that take a variable number of args,
    # a concept that doesn't exist in xpcom.
    def __repr__(self):
        iface_desc = self._get_classinfo_repr_()
        outer = _nsdom.IsOuterWindow(self._comobj_)
        return "<GlobalWindow outer=%s %s>" % (outer, iface_desc)

    # Open window/dialog
    def openDialog(self, url, name, features="", *args):
        svc = components.classes['@mozilla.org/embedcomp/window-watcher;1'] \
                .getService(components.interfaces.nsIWindowWatcher)
        # Wrap in an nsIArray with special support for Python being able to
        # access the raw objects if the call-site is smart enough
        args = _nsdom.MakeArray(args)
        ret = svc.openWindow(self._comobj_, url, name, features, args)
        # and re-wrap in one of our "dom" objects
        return _nsdom.MakeDOMObject(self._context_, ret)

    # window.open and window.openDialog seem identical except for *args???
    open = openDialog

    # Timeout related functions.
    def setTimeout(self, interval, handler, *args):
        return _nsdom.SetTimeoutOrInterval(self._comobj_, interval, handler,
                                           args, False)

    # setInterval appears to have reversed args???
    def setInterval(self, handler, interval, *args):
        return _nsdom.SetTimeoutOrInterval(self._comobj_, interval, handler,
                                           args, True)

    def clearTimeout(self, tid):
        return _nsdom.ClearTimeoutOrInterval(self._comobj_, tid)

    def clearInterval(self, tid):
        return _nsdom.ClearTimeoutOrInterval(self._comobj_, tid)

# Our "script context" - morally an nsIScriptContext, although that lives
# in our C++ code and delegates to this.
class ScriptContext:
    def __init__(self):
        self.globalNamespace = {} # must not change identity!
        self._remembered_objects_ = {} # could, but doesn't
        self._reset()

    def _reset(self):
        # Explicitly wipe all 'expandos' for our remembered objects.
        # Its not clear this is necessary, but it is easy to envisage someone
        # setting up a cycle via expandos.
        for ro in self._remembered_objects_.itervalues():
            ro._expandos_.clear()
        self._remembered_objects_.clear()
        # self.globalObject is the "outer window" global, whereas the
        # inner window global tends to be passed in to each function as the
        # 'scope' object.
        self.globalObject = None
        self.globalNamespace.clear()

    def __repr__(self):
        return "<ScriptContext at %d>" % id(self)

    # Called by the _nsdom C++ support to wrap the DOM objects.
    def MakeInterfaceResult(self, object, iid):
        if 0 and __debug__: # this is a little noisy...
            logger.debug("MakeInterfaceResult for %r (remembered=%s)",
                         object,
                         self._remembered_objects_.has_key(object))
        assert not hasattr(object, "_comobj_"), "should not be wrapped!"
        # If it is remembered, just return that object.
        try:
            return self._remembered_objects_[object]
        except KeyError:
            # Make a new wrapper - but don't remember the wrapper until
            # we need to, when a property is set on it.

            # We should probably QI for nsIClassInfo, and only do this special
            # wrapping for objects with the DOM flag set.
        
            if iid in GlobalWindowIIDs:
                klass = WrappedNativeGlobal
            else:
                klass = WrappedNative
            return klass(self, object, iid)

    # Called whenever this object must be "permanently" attached to the
    # context.  Typically this means an attribute or event handler
    # has been set on the object, which should "persist" while the context
    # is alive (ie, future requests to getElementById(), for example, must
    # return the same underlying object, with event handlers and properties
    # still in-place.
    def _remember_object(self, object):
        # You must only try and remember a wrapped object.
        # Once an object has been wrapped once, all further requests must
        # be identical
        assert self._remembered_objects_.get(object._comobj_, object)==object, \
               "Previously remembered object is not this object!"
        self._remembered_objects_[object._comobj_] = object
        if __debug__:
            logger.debug("%s remembering object %r - now %d items", self,
                         object, len(self._remembered_objects_))

    def _fixArg(self, arg):
        if arg is None:
            return []
        # Already a tuple?  This means the original Python args have been
        # found and passed directly.
        if type(arg) == types.TupleType:
            return arg
        try:
            argv = arg.QueryInterface(components.interfaces.nsIArray)
        except COMException, why:
            if why.errno != nsError.NS_NOINTERFACE:
                raise
            # This is not an array - see if it is a variant or primitive.
            try:
                var = arg.queryInterface(components.interfaces.nsIVariant)
                parent = None
                if self.globalObject is not None:
                    parent = self.globalObject._comobj_
                if parent is None:
                    logger.warning("_fixArg for context with no global??")
                return _xpcom.GetVariantValue(var, parent)
            except COMException, why:
                if why.errno != nsError.NS_NOINTERFACE:
                    raise
                try:
                    return primitives.GetPrimitive(arg)
                except COMException, why:
                    if why.errno != nsError.NS_NOINTERFACE:
                        raise
                return arg
        # Its an array - do each item
        ret = []
        for i in range(argv.length):
            val = argv.queryElementAt(i, components.interfaces.nsISupports)
            ret.append(self._fixArg(val))
        return ret

    def GetNativeGlobal(self):
        return self.globalNamespace

    def CreateNativeGlobalForInner(self, globalObject, isChrome):
        ret = dict()
        if __debug__:
            logger.debug("%r CreateNativeGlobalForInner returning %d",
                         self, id(ret))
        return ret

    def ConnectToInner(self, newInner, globalScope, innerScope):
        if __debug__:
            logger.debug("%r ConnectToInner(%r, %d, %d)",
                         self, newInner, id(globalScope), id(innerScope))
        globalScope['_inner_'] = innerScope # will do for now.

        self._prepareGlobalNamespace(newInner, innerScope)

    def WillInitializeContext(self):
        if __debug__:
            logger.debug("%r WillInitializeContent", self)

    def DidInitializeContext(self):
        if __debug__:
            logger.debug("%r DidInitializeContent", self)

    def DidSetDocument(self, doc, scope):
        if __debug__:
            logger.debug("%r DidSetDocument doc=%r scope=%d",
                         self, doc, id(scope))
        scope['document'] = doc

    def _prepareGlobalNamespace(self, globalObject, globalNamespace):
        assert isinstance(globalObject, WrappedNativeGlobal), \
               "Our global should have been wrapped in WrappedNativeGlobal"
        if __debug__:
            logger.debug("%r initializing (outer=%s), ns=%d", self,
                         _nsdom.IsOuterWindow(globalObject),
                         id(globalNamespace))
        assert len(globalObject._expandos_)==0, \
               "already initialized this global?"
        ns = globalObject.__dict__['_expandos_'] = globalNamespace
        self._remember_object(globalObject)
        ns['window'] = globalObject
        # we can't fetch 'document' and stash it now - there may not be one
        # at this instant (ie, it may be in the process of being attached)
        # Poke some other utilities etc into the namespace.
        ns['dump'] = dump

    def InitContext(self, globalObject):
        self._reset()
        self.globalObject = globalObject
        if globalObject is None:
            logger.debug("%r initializing with NULL global, ns=%d", self,
                         id(self.globalNamespace))
        else:
            self._prepareGlobalNamespace(globalObject, self.globalNamespace)

    def FinalizeClasses(self, globalObject):
        self._reset()

    def ClearScope(self, globalObject):
        if __debug__:
            logger.debug("%s.ClearScope (%d)", self, id(globalObject))
        globalObject.clear()

    def FinalizeContext(self):
        if __debug__:
            logger.debug("%s.FinalizeContext", self)
        self._reset()

    def EvaluateString(self, script, glob, principal, url, lineno, ver):
        # This appears misnamed - return value it not used.  Therefore we can
        # just do a simple 'exec'.  If we needed a return value, we would have
        # to treat this like a string event-handler, and compile as a temp
        # function.
        assert type(glob) == types.DictType
        # compile then exec to make use of lineno
        co = domcompile.compile(script, url, lineno=lineno-1)
        exec co in glob

    def ExecuteScript(self, scriptObject, scopeObject):
        if __debug__:
            logger.debug("%s.ExecuteScript %r in scope %s",
                         self, scriptObject, id(scopeObject))
        if scopeObject is None:
            scopeObject = self.GetNativeGlobal()
        assert type(scriptObject) == types.CodeType, \
               "Script object should be a code object (got %r)" % (scriptObject,)
        exec scriptObject in scopeObject

    def CompileScript(self, text, scopeObject, principal, url, lineno, version):
        # The line number passed is the first; offset is -1
        return domcompile.compile(text, url, lineno=lineno-1)

    def CompileEventHandler(self, name, argNames, body, url, lineno, version):
        if __debug__:
            logger.debug("%s.CompileEventHandler %s %s:%s ('%s')",
                         self, name, url, lineno, body[:100])
        co = domcompile.compile_function(body, url, name, argNames,
                                         lineno=lineno-1)
        g = {}
        exec co in g
        handler = g[name]
        # Flag that this function started life as inline script code,
        # so we later ensure the correct globals are used when the event fires
        handler._nsdom_compiled = True
        return g[name]

    def CallEventHandler(self, target, scope, handler, argv):
        if __debug__:
            logger.debug("CallEventHandler %r on target %s in scope %s",
                         handler, target, id(scope))
        # Event handlers must fire in their real globals (not a copy) so
        # it can set global vars!
        # If this was an inline event handler, re-bind to the window's globals.
        if hasattr(handler, '_nsdom_compiled'):
            globs = scope
            # re-bind to new globals
            f = new.function(handler.func_code, globs, handler.func_name,
                             handler.func_defaults)
            handler = f
        args = tuple(self._fixArg(argv))
        # We support having less args declared than supplied, a-la JS.
        args = args[:handler.func_code.co_argcount]
        return handler(*args)

    def BindCompiledEventHandler(self, target, scope, name, handler):
        if __debug__:
            logger.debug("%s.BindCompiledEventHandler (%s=%r) on target %s in scope %s",
                         self, name, handler, target, id(scope))
        # Keeps a ref to both the target and handler.
        self._remember_object(target)
        ns = target._expandos_
        ns[name] = handler

    def GetBoundEventHandler(self, target, scope, name):
        if __debug__:
            logger.debug("%s.GetBoundEventHandler for '%s' on target %s in scope %s",
                         self, name, target, id(scope))
        ns = target._expandos_
        return ns[name]

    def SetProperty(self, target, name, value):
        if __debug__:
            logger.debug("%s.SetProperty for %s=%r", self, name, value)
        target[name] = self._fixArg(value)
